#!@BIN_PERL@

# nagios: -epn
# disable embedded perl, works only under 3.x
# to avoid not preserving NAGIOS_* environment variables
#

#
#    Short description: Creates/Updates RRA archives, from nagios performance data
#    Copyright (C) 2006 - 2007  Badri Prasad Pillai
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

#
#    Contact details:
#		Badri(at)DigLinks.COM
#

use strict;

use RRDs;
use Getopt::Std;

sub log_msg($);
sub load_config($);
sub usage();
sub list_default_values();

use vars qw(
      $version
      $opt
      $rrderror
      %config
      @parse_service_str
      $service
      @data
      $ds_name
      $ds_value
      %ds_rewrite
      $rrd_file
      $nagios_env_file
      $perf_rrd_file
      );

$version = "1.3.2";
$opt      = {};
$rrderror = "";

#
# Do not buffer writes
$| = 1;

getopts( "c:de:l:LhH:T:o:s:", $opt );


#
# If no configuration file given, then use default values
#
%config = (
	CONF_DIR => "/etc/n2rrd",
	TEMPLATES_DIR => "templates",
	SERVICE_NAME_MAPS => "templates/maps/service_name_maps",
	ZOOM_JS => "js/zoom.js",
	RRA_DIR  => "/var/log/monitor/n2rrd/rra",
	LOGFILE => "/var/log/monitor/n2rrd/rra/n2rrd.log",
	DOCUMENT_ROOT => "/var/www/html",
	CACHE_DIR	=> "rrd-images",
	RRDTOOL => "/usr/bin/rrdtool",
	RRD_PATH_HIDDEN => 0,
	NAGIOS_HOST_URL => 1,
	THUMB_WIDTH => 200,
	THUMB_HEIGHT => 100,
	THUMB_DISPLAY => "Daily",
	THUMB_DISPLAY_COLUMNS => 3,
	SAVE_NAGIOS_ENVIRONMENT_VARS => 0,
	TMPDIR => "/tmp",
	CGIBIN => "cgi-bin",
	NAGIOS_CGIBIN => "nagios/cgi-bin",
	TEMPLATE_NAME_AT_BEGINNING => 0,
);

list_default_values if $opt->{L};

usage if $opt->{h};

load_config($opt->{c}) if $opt->{c};

#
# Sanity check: if RRA directory exists
log_msg "Direcotry rra_dir = \"$config{CONF_DIR}\" doesn't exists\n"
  and exit 1
  if !-d $config{CONF_DIR};

#
# Sanity check: if template directory exists
log_msg "Directory templates_dir = \"$config{TEMPLATES_DIR}\" doesn't exists\n"
  and exit 1
  if !-d "$config{CONF_DIR}/$config{TEMPLATES_DIR}";

#
# All three arguments are required
#
usage if !$opt->{H};
usage if !$opt->{s};
usage if !$opt->{o};

if ( $opt->{d} ) {
    my $s =
      "Host = $opt->{H}, Service name = $opt->{s}, Check result = $opt->{o}\n";
    log_msg $s if $opt->{d};
}

# if you use _ in your service name and you don't want cut the name, you can simulate a rewrite

$service = "";

open S_MAPS, '<', "$config{CONF_DIR}/$config{SERVICE_NAME_MAPS}"
      or log_msg "Can't open service maps file \"$config{CONF_DIR}/$config{SERVICE_NAME_MAPS}\"\n"
      and exit 1;
log_msg
"..Searching map in file \"$config{CONF_DIR}/$config{SERVICE_NAME_MAPS}\" for service \"$opt->{s}\" \n"
      if $opt->{d};

while (<S_MAPS>) {
        next if /^#/;    # Skip comments
        next if /^$/;    # Skip empty lines
        s/#.*//;         # Remove partial comments
        chomp;

        if ( /$opt->{s}:\s+(\S+)/i ) {
          $service = $1 if /$opt->{s}:\s+(\S+)/i;
          log_msg "...Mapping service $opt->{s} -> $service\n" if $opt->{d};
        }
}
close S_MAPS;


if ( !$service ) {
        # NO Mapping Found

        # assume delimiter between service name is '_'
        @parse_service_str = split /_/, $opt->{s};

        if ($config{TEMPLATE_NAME_AT_BEGINNING}) {
          # The first word after terminator character "_"
          $service = $parse_service_str[0];
        }
        else {
          # The last word after terminator character "_"
          $service = $parse_service_str[$#parse_service_str];
        }
}


#
# assume delimiter between ds_name and ds_value is '='

@data = ();

# possibility for external performance data parsing
if ( -f "$config{CONF_DIR}/$config{TEMPLATES_DIR}/code/${service}.pl" ) {
    open P, '<', "$config{CONF_DIR}/$config{TEMPLATES_DIR}/code/${service}.pl" 
		or die "Can't open perl code file \"$config{CONF_DIR}/$config{TEMPLATES_DIR}/code/${service}.pl\"";

    my @PCODE = <P>;
    close P;

    log_msg "Running perl code for host \"$opt->{H}\" and service \"$opt->{s}($service)\"\n" if $opt->{d};
    #
    # For external code uses nagios variables to get the value
    my $ret_str = eval join("\n",@PCODE);warn $@ if $@;

    log_msg "Perl code result for host \"$opt->{H}\" and service \"$service\" returned string \"$ret_str\"\n" 
		                                              if $opt->{d};
    #
    # remove spaces before and/or after "=" character
    $ret_str =~ s/\s+=/=/g;
    $ret_str =~ s/=\s+/=/g;


    @data = split /\s/, $ret_str;
}
else {

    #
    # remove spaces before and/or after "=" character
    $opt->{o} =~ s/\s+=/=/g;
    $opt->{o} =~ s/=\s+/=/g;
    @data = ( $opt->{o} =~ /'[\w\s]+'=\-?[0-9\.]+|\S+=\-?[0-9\.]+/g );
}

log_msg "Missing ds_name in Output: $opt->{o}, for service $opt->{s}\n" and exit 1 if !@data;

$ds_name  = "";
$ds_value = "";

foreach my $kv (@data) {
    my ( $key, $val ) = split /=/, $kv;
    $ds_name .= "$key:" if $key;
    $ds_value .= "$val:";
}
$ds_name  =~ s/:$//;
$ds_value =~ s/:$//;

%ds_rewrite = ();

$rrd_file = "$config{RRA_DIR}/$opt->{H}/$opt->{H}_${service}.rrd";
$perf_rrd_file = "$config{RRA_DIR}/$opt->{H}/$opt->{H}_${service}_el.rrd";

#
# create host dir if it doesn't exist
mkdir "$config{RRA_DIR}/$opt->{H}" if ! -d "$config{RRA_DIR}/$opt->{H}";

#
# Check if a rewrite rules exist
#
if ( -f "$config{CONF_DIR}/$config{TEMPLATES_DIR}/rewrite/$opt->{H}_${service}_rewrite" ) {

    #
    #	print "Rewrite file detected for: $opt->{H}_${service}\n";
    open REWRITE, '<', "$config{CONF_DIR}/$config{TEMPLATES_DIR}/rewrite/$opt->{H}_${service}_rewrite"
      or log_msg "Can't open rewrite rules file for $opt->{H}\n"
      and exit 1;

    my $ds_name_tmp = "";

    while (<REWRITE>) {
        next if /^#/;    # Skip comments
        next if /^$/;    # Skip empty lines
        s/#.*//;         # Remove partial comments
        chomp;

        $rrd_file      = $1 if /^rrd_file\s+(\S+)/;
        $perf_rrd_file = $1 if /^perf_rrd_file\s+(\S+)/;
        $ds_name_tmp .= "$2:" and $ds_rewrite{$1} = $2
          if /^ds_name\s+('[\w\s]+'|\S+)\s+(\S+)/;
    }
    close REWRITE;

    #print "rrd_file = $rrd_file\n";
    if ($ds_name_tmp) {
        $ds_name_tmp =~ s/:$//;
        $ds_name = $ds_name_tmp;
    }
}

$nagios_env_file = $rrd_file;
$nagios_env_file =~ s/\.rrd//;
$nagios_env_file .= ".env";

#
# if SAVE_NAGIOS_ENVIRONMENT_VARS is disabled
# remove env files if they exists
#
unlink $nagios_env_file if -f $nagios_env_file;

if ($config{SAVE_NAGIOS_ENVIRONMENT_VARS}) {
  # save NAGIOS environemnt values
  open E, '>', "$nagios_env_file"
      or log_msg "can't open file \"$nagios_env_file\"\n"
             and exit 1;
  log_msg "Writing ENV variables to file $nagios_env_file\n" if $opt->{d};
  foreach my $e (keys %ENV) {
     next if $e !~ /^NAGIOS/;
     print E "${e}=${ENV{$e}}\n" if length ($ENV{$e}) > 1;
  }
  close E;
}
#
# Create/Update 
#
if ( -e $rrd_file ) {

    #
    # If not time given
    my $time = $opt->{T} || "N";

    log_msg "Updating ds_names: $ds_name, ds_values: $ds_value in file: $rrd_file\n" if $opt->{d};

    RRDs::update( "$rrd_file", "--template", "$ds_name", "$time:$ds_value" );
    $rrderror = RRDs::error;
    log_msg "ERROR while updating $rrd_file: $rrderror\n" and exit 1
      if $rrderror;
}
else {

    #
    # First check for template with hostname
    my $rrd_create_template =
      "$config{CONF_DIR}/$config{TEMPLATES_DIR}/rra/$opt->{H}_${service}.t";

    if ( !-f $rrd_create_template ) {

        #
        # Else look for default template
        $rrd_create_template = "$config{CONF_DIR}/$config{TEMPLATES_DIR}/rra/${service}.t";
        if ( !-f $rrd_create_template ) {
            log_msg "Missing template for \"$opt->{H}\" service \"$opt->{s}\" \"$rrd_create_template\"\n";
            exit 1;
        }
    }

    log_msg "Template for service \"${service}\" -> \"$rrd_create_template\"\n" if $opt->{d};

    open T, '<', "$rrd_create_template"
      or log_msg "can't open file \"$rrd_create_template\"\n"
      and exit 1;

    my @t_params = ();

    while (<T>) {
        next if /^#/;    # Skip comments
        next if /^$/;    # Skip empty lines
        s/#.*//;         # Remove partial comments

        chomp;

        #
        # if there are ds_name rewrites needed
        foreach my $k ( keys %ds_rewrite ) {
            s/:$k:/:$ds_rewrite{$k}:/;
        }

        push @t_params, "$_";
    }

    close T;

    log_msg "Creating RRD file: $rrd_file\n" if $opt->{d};
    RRDs::create( $rrd_file, @t_params );
    $rrderror = RRDs::error;
    log_msg "Problem while creating rrd: $rrderror\n" and exit 1 if ($rrderror);
}

#
# Store execution and latency performance data
if ( $opt->{e} and $opt->{l} ) {
    my $time = $opt->{T} || "N";

    my $s =
"Host = $opt->{H}, Service name = $opt->{s}, Execution time = $opt->{e}, Latency = $opt->{l}\n";
    log_msg $s if $opt->{d};

    if ( -e $perf_rrd_file ) {
        log_msg "Updating execution and latency performance in file: $perf_rrd_file\n" if $opt->{d};
        RRDs::update(
            "$perf_rrd_file",    "--template",
            "execution:latency", "$time:$opt->{e}:$opt->{l}"
        );
        $rrderror = RRDs::error;
        log_msg
"ERROR while updating $perf_rrd_file: $rrderror e =$opt->{e} l = $opt->{l}\n"
          and exit 1
          if $rrderror;
    }
    else {

      #
      # First check for template with hostname and service name
      # it can be that a perticular service is polled at different interval than
      # default 500 seconds
      #
        my $perf_rrd_create_template =
          "$config{CONF_DIR}/$config{TEMPLATES_DIR}/rra/${service}_el.t";

        if ( !-f $perf_rrd_create_template ) {

            #
            # Else look for default template
            $perf_rrd_create_template =
              "$config{CONF_DIR}/$config{TEMPLATES_DIR}/rra/default_execution_latency_el.t";
            if ( !-f $perf_rrd_create_template ) {
                log_msg
"Missing execution and latency template \"$perf_rrd_create_template\"\n";
                exit 1;
            }
        }

        open EL, '<', "$perf_rrd_create_template"
          or log_msg "can't open file \"$perf_rrd_create_template\"\n"
          and exit 1;

        my @el_params = ();

        while (<EL>) {
            next if /^#/;    # Skip comments
            next if /^$/;    # Skip empty lines
            s/#.*//;         # Remove partial comments

            chomp;

            push @el_params, "$_";
        }
        close EL;

        log_msg
"Creating RRD execution and latency performance file: $perf_rrd_file\n"
          if $opt->{d};
        RRDs::create( $perf_rrd_file, @el_params );
        $rrderror = RRDs::error;
        log_msg "Problem while creating execution and latency rrd: $rrderror\n"
          and exit 1
          if ($rrderror);
    }

}


exit 0;

sub log_msg ($) {
    my $str = shift;

    open L, '>>', "$config{LOGFILE}"
      or die "Can't open logfile \"$config{LOGFILE}\"";

    open STDERR, '>&L';# dup STDERR

    print L "n2rrd:$opt->{H}: $str";

    close L;
}


sub load_config ($) {
	my $config_file = shift;
	my $line   = 0;

	open CONF, '<', "$config_file" 
	  or die "Can't open configuration file -> \"$config_file\":$!\n";
	
	
	#
	# Parse configuration file
	while (<CONF>) {
	    $line++;  # note confirguration file line number
	    chomp;    # Remove newline character
	    s/ //g;    # remove spaces
	    next if /^#/;    # Skip comments
	    next if /^$/;    # Skip empty lines
	    s/#.*//;         # Remove partial comments
	
	    if (/^(.*)=(.*)$/) {
		if (defined $config{$1}) {
	            $config{$1} = $2;
	        }
	        else {    # A warning should be ok
	            log_msg
	"WARNING: Unknown global variable declaration in line $line:\"$1 = $2\"\n";
	        }
	    }
	}
	close CONF;
}

sub usage () {
    print
"n2rrd: $version\n n2rrd options
	-h 		print usage and exit
	-c <path>/config-file-name 
			n2rrd.conf in case you want to overide default values
	-L 		this will print the default configuration values and exit
			
			pass following nagios 2.x variables as option, 1.x may differ
	-e Elapsed_time \$SERVICEEXECUTIONTIME\$
	-l Latence_time \$SERVICELATENCY\$
	-T time 	\$LASTHOSTCHECK\$
	-H hostname 	\$HOSTNAME\$ 
	-s servicename  \"\$SERVICEDESC\$\"
	-o service_check_output
			\"\$SERVICEPERFDATA\$\"\n\n";
    exit 0;
}

sub list_default_values() {
	load_config($opt->{c}) if $opt->{c};
	
	printf "#\n#n2rrd: $version\n";
        printf "#%24s = %s\n", "Key  ", "  Value";
        printf "%25s = %s\n", "#" x 25, "#" x 30;
	foreach my $k (sort keys %config) {
		printf "%25s = %s\n", $k, $config{$k};
	}
	exit 0;
}
